from path import Path
import numpy as np
import contextlib
import tempfile
import subprocess
import tarfile
import atexit
import shutil
import re
import os


def read_input_parameters(infile='pw.pwi'):
    """Return a dictionary of input arguments from an Espresso file.

    Parameters:
    -----------
    infile : str
        Input file to read arguments from.

    Returns:
    --------
    data : dict
        Arguments from an input file.
    """
    data = {}
    with open(infile) as f:
        lines = f.read().split('\n')

        for i, line in enumerate(lines):
            if '=' in line:
                key, value = [_.strip() for _ in line.split('=')]
                value = fortran_to_value(value)
                data[key] = value

            elif 'K_POINTS automatic' in line:
                kpts = np.array(lines[i + 1].split(), dtype=int)
                data['kpts'] = tuple(kpts[:3])
            elif 'K_POINTS gamma' in line:
                data['kpts'] = 1

    return data


def grepy(search, filename, instance=-1):
    """Perform a python based grep-like operation for a
    regular expression on a given file.

    Parameters:
    -----------
    search : str
        Regular expression to be found.
    filename : str
        File to be searched within.
    instances : slice
        Index of the arguments to find. If None, return all found results.

    Returns:
    --------
    results : list of str (N,) | None
        All requested instances of a given argument.
    """
    if not os.path.exists(filename):
        return None

    results = []
    with open(filename, 'r') as f:
        for line in f:
            if re.search(search, line, re.IGNORECASE):
                results += [line]

    if not results:
        return None

    if isinstance(instance, int):
        return results[instance]
    else:
        return results


@contextlib.contextmanager
def cd(path):
    """Does path management: if the path doesn't exists, create it
    otherwise, move into it until the indentation is broken.

    Parameters
    ----------
    path : str
        Directory path to create and change into.
    """
    cwd = os.getcwd()
    try:
        if not os.path.exists(path):
            os.makedirs(path)
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


def load_calculator(infile='calc.tar.gz', outdir='.'):
    """Unpack the contents of calc.save directory which
    was previously compressed.

    Parameters:
    -----------
    infile : str
        Relative input file for compressed save folder.
    outdir : str
        Relative output directory path for calc.save folder.
    """
    with tarfile.open(infile, 'r:gz') as f:
        f.extractall(outdir)


def value_to_fortran(value):
    """Return a Python compatible version of a FORTRAN argument.

    Parameters:
    -----------
    value : bool | int | float | str
        A Python friendly representation of the input value.

    Returns:
    --------
    fortran_value : str
        A FORTRAN argument.
    """
    if isinstance(value, bool):
        fortran_value = '.{}.'.format(str(value).lower())
    elif isinstance(value, float):
        fortran_value = str(value)
        if 'e' not in value:
            fortran_value += 'd0'

    return fortran_value


def fortran_to_value(fortran_value):
    """Return a Python compatible version of a FORTRAN argument.

    Parameters:
    -----------
    fortran_value : str
        A FORTRAN argument.

    Returns:
    --------
    value : bool | int | float | str
        A Python friendly representation of the input value.
    """
    if fortran_value == '.true.':
        return True
    elif fortran_value == '.false.':
        return False

    try:
        value = int(fortran_value)
    except(ValueError):
        try:
            value = float(fortran_value)
        except(ValueError):
            value = fortran_value.strip("'")

    return value


class SiteConfig():
    """Site configuration holding details about the execution environment
    with methods for retrieving the details from systems variables and
    creating directories.

    Parameters:
    -----------
    scheduler : str
        Name of the scheduler, curretly supports: SLURM, and
        PBS/TORQUE, and LSF
    """

    def __init__(self, scheduler=None, cluster=None):
        self.scheduler = scheduler
        self.cluster = cluster
        self.scratch = None
        self.nnodes = None
        self.nodelist = None
        self.nprocs = None

        if self.scheduler is None:
            self.submitdir = Path(os.getcwd())
            self.jobid = os.getpid()
            return

        lsheduler = self.scheduler.lower()
        if lsheduler == 'slurm':
            self.set_slurm_env()
        elif lsheduler in ['pbs', 'torque']:
            self.set_pbs_env()
        elif lsheduler == 'lbs':
            self.set_lbs_env()

    @classmethod
    def check_scheduler(cls):
        """Check for appropriate environment variables for the name
        of the cluster being used. Returns None for both if no
        supported schedule is found.

        Returns:
        --------
        cls : SiteConfig object
            Runs SiteConfig with the found arguments.
        """
        check_shedulers = {
            'SLURM': 'SLURM_CLUSTER_NAME',
            'PBS': 'PBS_SERVER',
            'LBS': 'LSB_EXEC_CLUSTER'}

        scheduler = None
        for sched, ev in check_shedulers.items():
            cluster = os.getenv(ev)
            if cluster:
                scheduler = sched
                break

        return cls(scheduler, cluster)

    def set_slurm_env(self):
        """Set the attributes necessary to run the job based on the
        enviromental variables associated with SLURM scheduler.
        """
        self.jobid = os.getenv('SLURM_JOB_ID')
        self.submitdir = Path(os.getenv('SLURM_SUBMIT_DIR'))

        tpn = int(os.getenv('SLURM_TASKS_PER_NODE').split('(')[0])

        cmd = ['scontrol', 'show', 'hostnames',
               os.getenv('SLURM_JOB_NODELIST')]
        self.nodelist = subprocess.check_output(cmd).decode().split('\n')[:-1]
        proclist = sorted(self.nodelist * tpn)

        self.nnodes = int(os.getenv('SLURM_JOB_NUM_NODES'))
        self.nprocs = len(proclist)

    def set_lbs_env(self):
        """Set the attributes necessary to run the job based on the
        enviromental variables associated with LBS scheduler.
        """
        self.jobid = os.getenv('LSB_BATCH_JID')
        self.submitdir = Path(os.getenv('LS_SUBCWD'))

        nodefile = os.getenv('LSB_HOSTS').split()
        procs = [_ for _ in nodefile]

        self.nodelist = list(set(procs))
        self.nnodes = len(self.nodelist)
        self.nprocs = len(procs)

    def set_pbs_env(self):
        """Set the attributes necessary to run the job based on the
        enviromental variables associated with PBS/TORQUE scheduler.
        """
        self.jobid = os.getenv('PBS_JOBID')
        self.submitdir = Path(os.getenv('PBS_O_WORKDIR'))

        nodefile = os.getenv('PBS_NODEFILE')
        with open(nodefile, 'r') as f:
            procs = [_.strip() for _ in f.readlines()]

        self.nodelist = list(set(procs))
        self.nnodes = len(self.nodelist)
        self.nprocs = len(procs)

    def make_scratch(self, save_calc=True):
        """Create a scratch directory on each calculation node if batch mode
        or a single scratch directory otherwise. Will attempt to call
        from $L_SCRATCH_JOB variable, then /tmp, then use the
        submission directory.

        This function will automatically cleanup upon exiting Python.

        Parameters:
        -----------
        save_calc : bool
            Whether to save the calculation folder or not.
        """
        scratch_paths = [os.getenv('ESPRESSO_SCRATCH', '/tmp'), self.submitdir]
        for scratch in scratch_paths:
            if os.path.exists(scratch):
                node_scratch = Path(scratch)
                break

        scratch = Path(tempfile.mkdtemp(
            prefix='qe_{}'.format(self.jobid), suffix='_scratch',
            dir=node_scratch)).abspath()

        with cd(self.submitdir):
            if self.scheduler:
                cmd = self.get_exe_command('mkdir -p {}'.format(scratch))
                subprocess.call(cmd)
            else:
                scratch.makedirs_p()

        self.scratch = scratch
        atexit.register(self.clean, save_calc)

    def get_exe_command(self, program, workdir=None):
        """Return a command as list to execute subprocess through
        a supplied argument. If a workdir is provided, assume
        execution per processor, otherwise, per host.

        Parameters:
        -----------
        program : str
            The full command line program to be executed using subprocess
        workdir : str
            A path to act as the working directory.

        Returns:
        --------
        command : list or str (N,)
            The list of arguments to be passed to subprocess.
        """
        exe, host, nproc, wd = ['mpiexec'], '-host', '-np', '-wdir'

        # SPECIAL SERVER EXECUTABLE CASES ARE HANDLED HERE
        if self.cluster == 'edison':
            exe, host, nproc, wd = ['srun'], '-w', '-n', '-D'
        elif self.cluster == 'slac' and workdir:
            exe = ['pam', '-g', '/afs/slac/g/suncat/bin/suncat-tsmpirun',
                   '-x', 'LD_LIBRARY_PATH']
        elif self.cluster == 'slac':
            exe = ['mpiexec', '--mca', 'orte_rsh_agent',
                   '/afs/slac.stanford.edu/package/lsf/bin.slac/gmmpirun_lsgrun.sh']
            init = exe + ['ls', self.submitdir]
            subprocess.call(init)

        # This indicates per-processor MPI run
        if workdir:
            command = exe + [wd, workdir]
        # Otherwise, per-host MPI run
        else:
            command = exe + [nproc, str(self.nnodes),
                             host, ','.join(self.nodelist)]
        command += program.split()

        return command

    def run(self, exe='pw.x', infile='pw.pwi', outfile='pw.pwo'):
        """Run an Espresso executable with subprocess. The executable will
        attempt to automatically assign an intelligent npool value
        for efficient parallelization.

        Parameters:
        -----------
        exe : str
            The Espresso executable command to be run.
        infile : str
            Name of the input file to be used for the executable.
        outfile : str
            Name of the output file to be used for the executable.

        Returns:
        --------
        state : int
            The output state of the subprocess executed command.
        """
        mypath = os.path.abspath(os.path.dirname(__file__))
        exedir = subprocess.check_output(['which', exe]).decode()

        # Copy the input file to the scratch directory.
        inp = self.submitdir.joinpath(infile)
        inp.copy(self.scratch)

        if self.scheduler:
            # Automatically assign npool for parallelization
            parflags = ''
            kpts = read_input_parameters()['kpts']
            if self.nprocs > 1 and kpts > self.nprocs:
                parflags += '-npool {}'.format(self.nprocs)

            command = self.get_exe_command(
                '{} {} -in {}'.format(exe, parflags, infile), self.scratch)
        else:
            command = [exe, '-in', infile]

        title = '{:<14}: {}\n{:<14}: {}{:<14}: {}\n'.format(
            'python dir', mypath,
            'espresso dir', exedir,
            'executable', ' '.join(command))

        # This will remove the old output file by default.
        output = self.submitdir.joinpath(outfile)
        with open(output, 'w') as f:
            f.write(title)

        with cd(self.scratch):
            with open(output, 'ab') as f:
                state = subprocess.call(command, stdout=f)

        if state != 0:
            if grepy('JOB DONE.', outfile):
                pass
            elif grepy('is really the minimum energy structure', outfile):
                pass
            else:
                # Read the error message
                error_message = []
                with open(outfile, 'r') as f:
                    for line in f:
                        # Capture the message between %%%
                        if '%%%%%%%%%%%%%%' in line:
                            error_message += [line]
                            if len(error_message) > 1:
                                break
                        elif error_message:
                            error_message += [line]

                raise RuntimeError('pw.x returned a nonzero exit state:\n'
                                   '{}'.format(''.join(error_message)))

        return state

    def clean(self, save_calc=True):
        """Remove temporary files and directories."""
        os.chdir(self.submitdir)

        calc = self.scratch.joinpath('calc.save')
        save = self.submitdir.joinpath('calc.tar.gz')

        if os.path.exists(calc) and save_calc:
            with tarfile.open(save, 'w:gz') as f:
                f.add(calc, arcname=calc.basename())

        shutil.rmtree(self.scratch)



from .espresso import Espresso
from . import siteconfig
from . import validate
import numpy as np
from ase.io import read
import glob


class PDOS(Espresso):
    """Calculate a projection of wavefunctions for a finished Quantum
    Espresso calculation.

    Assumes that the command is being run in the same directory as a
    completed .pwo file. The atoms object will be read in automatically.
    A calc.tar.gz file must be present for revisited runs.

    Calculation parameters are assumed to be unchanged unless otherwise
    specified in the keyword arguments. If any arguments specific to the
    original files input are changed, such as k-points, these values will
    be automatically updated and an NSCF calculation performed automatically.
    """

    def __init__(self, **kwargs):
        """Read in the atoms object and fermi level from an existing
        calculation and then initialize the Espresso parent class.

        Separates standard input parameters from projection file
        inputs so both can be read from keyword arguments simultaneously.
        Will ignore identical parameters.
        """
        self.atoms = read('pw.pwo')
        self.efermi = self.get_fermi_level()
        super().__init__(self.atoms, **kwargs)

        # Extract the PROJWFC variables
        projwfc_args = {}
        initial_params = self.params.copy()
        for key, value in initial_params.items():
            if key in validate.projwfc_vars:
                projwfc_args[key] = value
                del kwargs[key]

        self.projwfc_args = projwfc_args
        self.params = siteconfig.read_input_parameters()

        # Check for variables that have changed
        self.recalculate = False
        for parameter, value in kwargs.items():
            origional = self.params.get(parameter)

            if origional is not None:
                if isinstance(origional, float):
                    match = np.isclose(origional, value)
                else:
                    match = origional == value

                if not match:
                    self.recalculate = True
                    self.params.update(kwargs)
                    break

    def write_projwfc_input(self, infile='projwfc.pwi'):
        """Write a projected wavefunction file.

        Parameters:
        -----------
        infile : str
            Input file to write to.
        """
        with open(infile, 'w') as f:
            f.write("&PROJWFC\n   {:16} = 'calc'\n   {:16} = '.'\n".format(
                'prefix', 'outdir'))

            for key, value in self.projwfc_args.items():
                value = siteconfig.value_to_fortran(value)
                f.write('   {:16} = {}\n'.format(key, value))
            f.write('/\n')

    def calculate(self):
        """Perform a projwfc.x calculation. If scratch file exists,
        assume an Espresso calculation has been run within the script
        already. Otherwise, looks for calc.tar.gz calculation from a
        previous run in this directory.
        """
        if not self.site.scratch:
            # Starting from completed calculation. Doesn't overwrite save.
            self.site.make_scratch(save_calc=False)
            savefile = self.site.submitdir.joinpath('calc.tar.gz')
            if savefile.exists():
                siteconfig.load_calculator(
                    outdir=self.site.scratch.abspath())
            else:
                raise RuntimeError(
                    'Can not execute a post process calculation '
                    'without a calc.tar.gz file')

        if self.recalculate:
            self.calculate_ncsf()
        self.write_projwfc_input('projwfc.pwi')
        self.site.run('projwfc.x', 'projwfc.pwi', 'projwfc.pwo')

    def calculate_ncsf(self):
        """If the pw.pwi inputs have changed, run a ncsf calculation."""
        self.params['calculation'] = 'nscf'
        self.write_pw_input('nscf.pwi')
        self.site.run(infile='nscf.pwi', outfile='nscf.pwo')
        self.efermi = self.get_fermi_level('nscf.pwo')

    def get_pdos(self):
        """Return the projected densities of states.

        TODO: Most of this is copied from ase-espresso. Not sure what it's
        all of how it works yet.
        """
        channels = {'s': 0, 'p': 1, 'd': 2, 'f': 3}
        output = self.site.scratch.joinpath('calc.pdos_tot')
        states = np.loadtxt(output)
        energies = states[:, 0] - self.efermi

        if len(states[0]) > 3:
            nspin = 2
            dos = [states[:, 1], states[:, 2]]
        else:
            nspin = 1
            dos = states[:, 1]

        # read in projections onto atomic orbitals
        pdos = [{} for _ in self.atoms]
        globstr = '{}/calc.pdos_atm*'.format(self.site.scratch)
        proj = sorted(glob.glob(globstr))

        for infile in proj:
            data = np.genfromtxt(infile)

            split = infile.split('#')
            iatom = int(split[1].split('(')[0]) - 1
            channel = split[2].split('(')[1].rstrip(')').replace('_j', ',j=')
            jpos = channel.find('j=')

            if jpos < 0:
                # ncomponents = 2*l+1+1  (latter for m summed up)
                ncomponents = (2 * channels[channel[0]] + 2) * nspin
            else:
                # ncomponents = 2*j+1+1  (latter for m summed up)
                ncomponents = int(2 * float(channel[jpos + 2:])) + 2

            if channel not in list(pdos[iatom].keys()):
                pdos[iatom][channel] = np.zeros((ncomponents, len(energies)))

                for i in range(ncomponents):
                    pdos[iatom][channel][i] = data[:, (i + 1)]

        return energies, dos, pdos



from ase.units import Rydberg, Bohr
import numpy as np
import os
import warnings
import six


def dipfield(calc, val):
    """"""
    assert isinstance(val, bool)
    if not calc.get_param('tefield'):
        calc.params['tefield'] = True


def tefield(calc, val):
    """"""
    assert isinstance(val, bool)


def tstress(calc, val):
    """"""
    assert isinstance(val, bool)
    if not calc.get_param('tprnfor'):
        calc.params['tprnfor'] = True


def tprnfor(calc, val):
    """"""
    assert isinstance(val, bool)


def occupations(calc, val):
    """"""
    assert isinstance(val, six.string_types)
    values = ['smearing', 'tetrahedra', 'tetrahedra_lin',
              'tetrahedra_opt', 'fixed', 'from_input']

    assert val in values


def degauss(calc, val):
    """"""
    assert isinstance(val, (float, int))
    val /= Rydberg

    # Without smearing, we enforce fixed occupation.
    if abs(val) <= 1e-13:
        calc.params['occupations'] = 'fixed'
        moments = calc.atoms.get_initial_magnetic_moments()
        calc.params['tot_magnetization'] = moments.sum()
        calc.params['nspin'] = 2
        warnings.warn("'degauss' is zero. Enforcing fixed "
                      "overall magnetic moment")

    return val


def lspinorb(calc, val):
    """"""
    assert isinstance(val, bool)
    assert calc.get_param('noncolin')


def nspin(calc, val):
    """"""
    assert val in [2, 4]
    assert calc.get_param('occupations')
    assert calc.get_param('smearing')
    assert calc.get_param('degauss')
    if val == 4:
        assert calc.get_param('noncolin')

    moments = calc.atoms.get_initial_magnetic_moments()
    assert moments.any()


def noncolin(calc, val):
    """"""
    assert isinstance(val, bool)
    calc.params['nspin'] = 4
    nspin(calc, 4)


def tot_magnetization(calc, val):
    """"""
    assert isinstance(val, int)
    nspin(calc, calc.get_param('nspin'))


def ion_dynamics(calc, val):
    """"""
    assert isinstance(val, six.string_types)

    calculator = calc.get_param('calculation')
    calculation(calc, calculator)

    if calculator == 'relax':
        if calc.atoms.constraints is not None:
            values = ['damp']
        else:
            values = ['bfgs', 'damp']

        assert val in values

    elif calculator == 'md':
        if calc.atoms.constraints is not None:
            values = ['verlet']
        else:
            values = ['verlet', 'langevin', 'langevin-smc']

        assert val in values

    elif calculator == 'vc-relax':
        values = ['bfgs', 'damp']
        assert val in values

    elif calculator == 'vc-md':
        values = ['beeman']
        assert val in values

    else:
        raise ValueError("ion_dynamics requires calculator to be "
                         "'relax', 'md', 'vc-relax', or 'vc-md'")


def nbnd(calc, val):
    """If nbands is negative, assign additional bands."""
    assert isinstance(val, int)

    if val < 0:
        nvalence, nel = calc.get_nvalence()

        if calc.get_param('noncolin'):
            val = nvalence.sum() - val * 2
        else:
            val = int(nvalence.sum() / 2) - val

    return val


def etot_conv_thr(calc, val):
    """"""
    assert isinstance(val, (float, int))
    val /= Rydberg

    return val


def forc_conv_thr(calc, val):
    """"""
    assert isinstance(val, (float, int))
    val /= Rydberg / Bohr

    return val


def ecutwfc(calc, val):
    """"""
    assert isinstance(val, (float, int))
    val /= Rydberg

    return val


def ecutrho(calc, val):
    """"""
    assert isinstance(val, (float, int))
    val /= Rydberg

    return val


def conv_thr(calc, val):
    """"""
    assert isinstance(val, (float, int))
    val /= Rydberg

    return val


def ecutfock(calc, val):
    """"""
    assert isinstance(val, (float, int))
    val /= Rydberg

    return val


def calculation(calc, val):
    """"""
    assert isinstance(val, six.string_types)

    values = ['scf', 'nscf', 'bands', 'relax', 'md', 'vc-relax', 'vc-md']
    assert val in values


def prefix(calc, val):
    """"""
    assert isinstance(val, six.string_types)
    warnings.warn("For directory consistency, 'prefix' is ignored")


def outdir(calc, val):
    """"""
    assert isinstance(val, six.string_types)


def disk_io(calc, val):
    """"""
    assert isinstance(val, six.string_types)

    if isinstance(val, str):
        values = ['high', 'medium', 'low', 'none']
        assert val in values


def kpts(calc, val):
    """Test k-points to be 'gamma' or list_like of 3 values.
    Only automatic assignment is currently supported.

    https://www.quantum-espresso.org/Doc/INPUT_PW.html#idm45922794051696
    """
    if val == 'gamma':
        return
    assert isinstance(val, (tuple, list, np.ndarray))
    assert len(val) == 3


def input_dft(calc, val):
    """"""
    assert isinstance(val, six.string_types)


def mixing_beta(calc, val):
    """"""
    assert isinstance(val, float)
    assert (val > 0) and (val < 1)


def nosym(calc, val):
    """"""
    assert isinstance(val, bool)


# PROJWFC variables
def Emin(calc, val):
    """"""
    assert isinstance(val, (float, int))

    efermi = calc.get_fermi_level()
    if efermi is not None:
        return val + efermi


def Emax(calc, val):
    """"""
    assert isinstance(val, (float, int))

    efermi = calc.get_fermi_level()
    if efermi is not None:
        return val + efermi


def DeltaE(calc, val):
    """"""
    assert isinstance(val, (float, int))


def nguass(calc, val):
    """"""
    assert isinstance(val, int)


variables = {
    # CONTROL
    'calculation': 'scf',
    'title': None,
    'verbosity': None,
    'restart_mode': None,
    'wf_collect': None,
    'nstep': None,
    'iprint': None,
    'tstress': True,
    'tprnfor': True,
    'dt': None,
    'outdir': '.',
    'wfcdir': None,
    'prefix': 'calc',
    'lkpoint_dir': None,
    'max_seconds': None,
    'etot_conv_thr': 0.0,
    'forc_conv_thr': 0.05 / (Rydberg / Bohr),
    'disk_io': None,
    'pseudo_dir': os.environ['ESP_PSP_PATH'],
    'tefield': None,
    'dipfield': None,
    'lelfield': None,
    'nberrycyc': None,
    'lorbm': None,
    'lberry': None,
    'gdir': None,
    'nppstr': None,
    'lfcpopt': None,
    'gate': None,
    # SYSTEM
    'ibrav': 0,
    'celldm': None,
    'A': None,
    'B': None,
    'C': None,
    'cosAB': None,
    'cosAC': None,
    'cosBC': None,
    'nat': None,
    'ntyp': None,
    'nbnd': None,
    'tot_charge': None,
    'starting_charge': None,
    'tot_magnetization': None,
    'starting_magnetization': None,
    'ecutwfc': None,
    'ecutrho': None,
    'ecutfock': None,
    'nr1': None,
    'nr2': None,
    'nr3': None,
    'nr1s': None,
    'nr2s': None,
    'nr3s': None,
    'nosym': None,
    'nosym_evc': None,
    'noinv': None,
    'no_t_rev': None,
    'force_symmorphic': None,
    'use_all_frac': None,
    'occupations': 'smearing',
    'one_atom_occupations': None,
    'starting_spin_angle': None,
    'degauss': 0.1 / Rydberg,
    'smearing': 'fd',
    'nspin': None,
    'noncolin': None,
    'ecfixed': None,
    'qcutz': None,
    'q2sigma': None,
    'input_dft': None,
    'exx_fraction': None,
    'screening_parameter': None,
    'exxdiv_treatment': None,
    'x_gamma_extrapolation': None,
    'ecutvcut': None,
    'nqx1': None,
    'nqx2': None,
    'nqx3': None,
    'lda_plus_u': None,
    'lda_plus_u_kind': None,
    'Hubbard_U': None,
    'Hubbard_J0': None,
    'Hubbard_alpha': None,
    'Hubbard_beta': None,
    'Hubbard_J': None,
    'starting_ns_eigenvalue': None,
    'U_projection_type': None,
    'edir': None,
    'emaxpos': None,
    'eopreg': None,
    'eamp': None,
    'angle1': None,
    'angle2': None,
    'constrained_magnetization': None,
    'fixed_magnetization': None,
    'lambda': None,
    'report': None,
    'lspinorb': None,
    'assume_isolated': None,
    'esm_bc': None,
    'esm_w': None,
    'esm_efield': None,
    'esm_nfit': None,
    'fcp_mu': None,
    'vdw_corr': None,
    'london': None,
    'london_s6': None,
    'london_c6': None,
    'london_rvdw': None,
    'london_rcut': None,
    'ts_vdw_econv_thr': None,
    'ts_vdw_isolated': None,
    'xdm': None,
    'xdm_a1': None,
    'xdm_a2': None,
    'space_group': None,
    'uniqueb': None,
    'origin_choice': None,
    'rhombohedral': None,
    'zgate': None,
    'realxz': None,
    'block': None,
    'block_1': None,
    'block_2': None,
    'block_height': None,
    # ELECTRONS
    'electron_maxstep': None,
    'scf_must_converge': None,
    'conv_thr': None,
    'adaptive_thr': None,
    'conv_thr_init': None,
    'conv_thr_multi': None,
    'mixing_mode': None,
    'mixing_beta': None,
    'mixing_ndim': None,
    'mixing_fixed_ns': None,
    'diagonalization': None,
    'ortho_para': None,  # OBSOLETE
    'diago_thr_init': None,
    'diago_cg_maxiter': None,
    'diago_david_ndim': None,
    'diago_full_acc': None,
    'efield': None,
    'efield_cart': None,
    'efield_phase': None,
    'startingpot': None,
    'startingwfc': None,
    'tqr': None,
    # IONS
    'ion_dynamics': None,
    'ion_positions': None,
    'pot_extrapolation': None,
    'wfc_extrapolation': None,
    'remove_rigid_rot': None,
    'ion_temperature': None,
    'tempw': None,
    'tolp': None,
    'delta_t': None,
    'nraise': None,
    'refold_pos': None,
    'upscale': None,
    'bfgs_ndim': None,
    'trust_radius_max': None,
    'trust_radius_min': None,
    'trust_radius_ini': None,
    'w_1': None,
    'w_2': None,
    # CELL
    'cell_dynamics': None,
    'press': None,
    'wmass': None,
    'cell_factor': None,
    'press_conv_thr': None,
    'cell_dofree': None,
}


projwfc_vars = {
    # PROJWFC
    'ngauss': None,
    'Emin': None,
    'Emax': None,
    'DeltaE': None,
    'lsym': None,
    'pawpraj': None,
    'filpdos': None,
    'filproj': None,
    'lwrite_overlaps': None,
    'lbinary_data': None,
    'kresolveddos': None,
    'tdosinboxes': None,
    'n_proj_boxes': None,
    'irmin': None,
    'irmax': None,
    'plotboxes': None}



from . import siteconfig
from . import validate
import numpy as np
import six
import ase
import warnings
import os


class Espresso(ase.calculators.calculator.FileIOCalculator):
    """Decaf Espresso - A Light weight ASE interface for Quantum Espresso.

    Validation of arguments present in validate.py will be performed
    automatically.

    Configuration of site executables and scratch directories will be
    performed automatically with siteconfig.py
    """

    implemented_properties = [
        'energy', 'forces', 'stress', 'magmom', 'magmoms']

    def __init__(
            self,
            atoms=None,
            **kwargs):
        """Initialize the calculators by validating the input key word
        arguments corresponding to valid Quantum Espresso arguments:
        https://www.quantum-espresso.org/Doc/INPUT_PW.html

        Geometric aspects of the calculation are defined by the atoms
        object, so geometry related arguments are ignored.

        Parameters:
        -----------
        atoms : Atoms object
            ASE atoms to attach the calculator with.
        """
        self.params = kwargs.copy()
        self.defaults = validate.variables
        self.results = {}
        self.site = siteconfig.SiteConfig.check_scheduler()

        if atoms:
            atoms.set_calculator(self)
            self.symbols = self.atoms.get_chemical_symbols()
            self.species = np.unique(self.symbols)

        # Certain keys are used for fixed IO features or atoms object
        # information. For calculation consistency, user input is ignored.
        ignored_keys = ['prefix', 'outdir', 'ibrav', 'celldm', 'A', 'B', 'C',
                        'cosAB', 'cosAC', 'cosBC', 'nat', 'ntyp']

        # Run validation checks
        init_params = self.params.copy()
        for key, val in init_params.items():

            if key in ignored_keys:
                del self.params[key]
            elif key in validate.__dict__:
                f = validate.__dict__[key]
                new_val = f(self, val)

                # Used to convert values from eV to Ry
                if new_val is not None:
                    self.params[key] = new_val
                elif isinstance(val, six.string_types):
                    self.params[key] = str(val)
            else:
                warnings.warn('No validation for {}'.format(key))

    def write_pw_input(self, infile='pw.pwi'):
        """Create the input file to start the calculation. Defines unspecified
        defaults as defined in validate.py.

        Parameters:
        -----------
        infile : str
            Name of the input file which will be written.
        """
        for key, value in self.defaults.items():
            setting = self.get_param(key)
            if setting is not None:
                self.params[key] = setting

        # ASE format for the pseudopotential file location.
        self.params['pseudopotentials'] = {}
        for species in self.species:
            self.params['pseudopotentials'][species] = '{}.UPF'.format(species)

        ase.io.write(infile, self.atoms, **self.params)

    def calculate(self, atoms, properties=['energy'], changes=None):
        """Perform a calculation."""
        self.write_pw_input('pw.pwi')

        self.site.make_scratch()
        self.site.run(infile='pw.pwi', outfile='pw.pwo')

        atoms = ase.io.read('pw.pwo')
        self.set_atoms(atoms)
        self.set_results(atoms)

    def set_results(self, atoms):
        """Set the results of the calculator."""
        self.results = atoms._calc.results

    def set_atoms(self, atoms):
        """Set the atoms object to the calculator."""
        self.atoms = atoms.copy()

    def get_param(self, parameter):
        """Return the parameter associated with a calculator,
        otherwise, return the default value.

        Parameters:
        -----------
        parameter : str
            Name of the parameter to retrieve the value of.

        Returns:
        --------
        value : bool | int | float | str | None
            The parameter value specified by the user, or None
        """
        value = self.params.get(parameter, self.defaults[parameter])

        return value

    def get_nvalence(self):
        """Get number of valence electrons from pseudopotential file associated
        with an atoms object.

        Returns:
        --------
        nvalence : ndarray (N,)
            Number of valence electrons associated with each atom N.
        nel : dict
            Number of electrons associated with species in the atoms object.
        """
        if hasattr(self, 'nel'):
            return self.nvalence, self.nel

        nel = {}
        for species in self.species:
            fname = os.path.join(
                self.get_param('pseudo_dir'), '{}.UPF'.format(species))
            valence = siteconfig.grepy('z valence|z_valence', fname).split()[0]
            nel[species] = int(float(valence))

        nvalence = np.zeros_like(self.symbols, int)
        for i, symbol in enumerate(self.symbols):
            nvalence[i] = nel[symbol]

        # Store the results for nbnd validation.
        self.nvalence, self.nel = nvalence, nel

        return nvalence, nel

    @staticmethod
    def get_fermi_level(outfile='pw.pwo'):
        """Return the fermi level in eV from a completed calculation file.

        Parameters:
        -----------
        outfile : str
            The completed calculation file to read the fermi level from.

        Returns:
        --------
        efermi : float
            The fermi energy in eV.
        """
        efermi = siteconfig.grepy('Fermi energy', outfile)
        if efermi is not None:
            efermi = float(efermi.split()[-2])

        return efermi



import setuptools

with open('requirements.txt', 'r') as f:
    requirements = f.readlines()

with open('readme.org', 'r') as f:
    readme = f.read()

setuptools.setup(
    name="decaf-espresso",
    version="0.1.1",
    url="https://github.com/jboes/decaf-espresso",

    author="Jacob Boes",
    author_email="jacobboes@gmail.com",

    description="Light-weight ASE calculator wrapper for Quantum Espresso.",
    long_description=readme,
    license='GPL-3.0',

    packages=['espresso'],
    package_dir={'espresso': 'espresso'},
    install_requires=requirements,
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Chemistry',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
